imexp:= function( a, u )

        # a, u lie in the same Lie algebra, ad a is nilpotent
        # we return exp( ad a )(u)

        local v, u0, k;

        v:= u;
        u0:= u;
        k:=1;
        while not IsZero(u0) do
              u0:= a*u0/k;
              k:= k+1;
              v:= v+u0;
        od;

        return v;
end;


_exp:= function( u )

        # u is a nilpotent matrix, we compute exp(u)

        local u1, i, u2;

        if IsLieObject(u) then u:= u![1]; fi;
        
        u1:= u^0;
        i:=1;
        u2:= u;
        while not IsZero(u2) do
           u1:= u1+u2;
           u2:= u2*u/(i+1);
           i:= i+1;
        od;
        return u1;
end;

_log:= function( u )

        # u is a unipotent matrix, we compute log(u)

        local u1, u2, i, lgu, F;

        if IsLieObject(u) then u:= u![1]; fi;

        F:= DefaultFieldOfMatrix(u);
        u1:= u-u^0;
        u2:= u1;
        lgu:= Zero(F)*u;
        i:= 1;
        while not IsZero(u2) do
            lgu:= lgu+(-1)^(i-1)*u2/i;
            u2:= u2*u1; i:= i+1;
        od;

        return lgu;
end;


levidecomposition:= function( L, sigma )


        # here L is an algebraic matrix Lie algebra (char 0),
        # sigma is a conjugation of L
        # we compute a reductive subalgebra R of L such that L = R + N,
        # where N is the nilpotent radical. The output is [R,N].
        # All subalgebras are fixed under sigma.

        local ld, H, L1, done, sp, decs, bas, R, N, F, i, b0, x, u, T, K,
              ld0;


        if not ForAll( Basis(L), x -> x = sigma(x) ) then
           F:= LeftActingDomain(L);
           i:= E(4)*One( F );
           b0:= [ ];
           sp:= MutableBasis( F, [], Zero(L) );
           for x in Basis(L) do
               if sigma(x) = x then
                  if not IsContainedInSpan( sp, x ) then
                     Add( b0, x );
                     CloseMutableBasis( sp, x );
                  fi;
               else
                  u:= x+sigma(x);
                  if not IsContainedInSpan( sp, u ) then
                     Add( b0, u );
                     CloseMutableBasis( sp, u );
                  fi;
                  u:= i*(x-sigma(x));
                  if not IsContainedInSpan( sp, u ) then
                     Add( b0, u );
                     CloseMutableBasis( sp, u );
                  fi;
               fi;
           od;
           T:= StructureConstantsTable( Basis( L,b0 ) );
           K:= LieAlgebraByStructureConstants( F, T );
           ld0:= LeviMalcevDecomposition( K );
           ld:= [ ];
           Add( ld, Subalgebra( L,
              List( Basis(ld0[1]), x -> Coefficients(Basis(K),x)*b0 ) ) );
           Add( ld, Subalgebra( L,
              List( Basis(ld0[2]), x -> Coefficients(Basis(K),x)*b0 ) ) );
        else
           ld:= LeviMalcevDecomposition( L );
        fi;

        H:= CartanSubalgebra( ld[2] );

        L1:= ld[2];
        done:= false;
        while not done do
              sp:= ProductSpace( H, L1 );
              done:= sp = L1;
              if not done then L1:= sp; fi;
        od;

        decs:= List( Basis(H), x -> JordanDecomposition( x![1] ) );
        bas:= List( decs, x -> LieObject(x[1]) );
        bas:= Concatenation( Basis(ld[1]), Basis( Subalgebra( L, bas ) ) );

        if Length(bas)=0 then
           R:= TrivialSubalgebra(L);
        else
           R:= Subalgebra( L, bas, "basis" );
        fi;

        bas:= Concatenation( List( decs, x -> LieObject(x[2]) ), Basis(L1) );
        if Length(bas)=0 then
           N:= TrivialSubalgebra(L);
        else
           N:= Subalgebra( L, bas );
        fi;
        return [ R, N ];
end;
        
conjugatesemsimparts:= function( L, S1, S2, R )


        # here L is Lie algebra of char 0.
        # S1, S2 are two semisimple complements to the solvable radical N
        # We find a in [L,R] such that exp(ad a)(S1) = S2

        local DR, LR, rho, bas, B2, m, u, gam, i, j, k, eqns, rhs, eqn, sol,
              C, f, Lhat, S1hat, S2hat, Rhat, a, b, v, u0, K;

        if S1 = S2 then return Zero(R); fi;

        DR:= LieDerivedSubalgebra(R);
        LR:= ProductSpace( L, R );
        LR:= Subalgebra( L, BasisVectors(Basis(LR)) );

        if Dimension(DR)=0 and LR=R then

           rho:=[ ];
           bas:= Concatenation( Basis(S2), Basis(R) );
           B2:= Basis(L,bas);
           m:= Dimension(S1);
           for i in [1..m] do
               u:= Coefficients( B2, Basis(S1)[i] );
               u:= u{[m+1..m+Dimension(R)]};
               Add( rho, -u );
           od;

           gam:= List([1..Dimension(R)], i -> [] );
           for i in [1..Dimension(R)] do
               for j in [1..Dimension(S1)] do
                   u:= Coefficients( Basis(R), Basis(R)[i]*Basis(S1)[j] );
                   gam[i][j]:= u;
               od;
           od;

           eqns:= [ ]; rhs:= [ ];
           for i in [1..m] do
               for k in [1..Dimension(R)] do
                   eqn:= [ ];
                   for j in [1..Dimension(R)] do
                       Add( eqn, gam[j][i][k] );
                       Add( eqns, eqn );
                       Add( rhs, rho[i][k] );
                   od;
               od;
           od;

           sol:= SolutionMat( TransposedMat(eqns), rhs );
           return sol*Basis(R);

        else

           C:= LieCentre( LR );
           f:=NaturalHomomorphismByIdeal( L, C );

           Lhat:= Range( f );
           S1hat:= Subalgebra( Lhat, List( Basis(S1), x -> Image(f,x) ) );
           S2hat:= Subalgebra( Lhat, List( Basis(S2), x -> Image(f,x) ) );
           Rhat:= Subalgebra( Lhat, List( Basis(R), x -> Image(f,x) ) );
           a:= PreImagesRepresentative( f,
                        conjugatesemsimparts( Lhat, S1hat, S2hat, Rhat ) );

           K:= Subalgebra( L, Concatenation( Basis(S2), Basis(C) ) );

           bas:= [ ];
           for u in Basis(S1) do

               v:= u;
               u0:= u;
               k:=1;
               while not IsZero(u0) do
                  u0:= a*u0/k;
                  k:= k+1;
                  v:= v+u0;
               od;

               Add( bas, v );
           od;

           b:= conjugatesemsimparts( K, Subalgebra(K,bas), S2, C );
           return a+b;

        fi;

end;


conjCSAsolv:= function( L, H1, H2 )


        # here L is a solvable Lie algebra of characteristic 0,
        # H1, H2 Cartan subalgebras; the output is a list of elements
        # x1,...,xs of [L,L] such that
        # exp(ad x1)...exp(ad xs)(H1) = H2

        local ds, I, K, L1, done, sp, IisL1, f, Lhat, H1hat, H2hat,
              aa, len, bas, i, H0, bb, Om, cf, u, adun, k, B, y, z;

        if H1 = H2 then return [ ]; fi;

        ds:= LieDerivedSeries( L );
        I:= ds[ Length(ds)-1 ];

        K:= Subalgebra( L, Concatenation( Basis(H2), Basis(I) ) );
        if Dimension(K) = Dimension(L) then

           L1:= L;
           done:= false;
           while not done do
               sp:= ProductSpace( H2, L1 );
               if Dimension( sp ) < Dimension(L1) then
                  L1:= sp;
               else
                  done:= true;
               fi;
           od;

           I:= Ideal(L, Basis(L1) );
           IisL1:= true;
        else
           IisL1:= false;
        fi;

        f:= NaturalHomomorphismByIdeal( L, I );
        Lhat:= Range(f);
        H1hat:= Subalgebra( Lhat, List( Basis(H1), x -> Image( f, x ) ) );
        H2hat:= Subalgebra( Lhat, List( Basis(H2), x -> Image( f, x ) ) );

        aa:= List( conjCSAsolv( Lhat, H1hat, H2hat ), x ->
            PreImagesRepresentative( f, x ) );
        len:= Length(aa);
        bas:= BasisVectors( Basis(H1) );
        for i in [len,len-1..1] do
            bas:= List( bas, x -> imexp( aa[i], x ) );
        od;

        H0:= Subalgebra( L, bas );

        if not IisL1 then
           bb:= conjCSAsolv( K, H0, H2 );
           return Concatenation( bb, aa );

        else

           # find a possibly nice regular element in H2
           Om:= [0..Dimension(L)];
           done:= false;
           while not done do
              cf:= List( [1..Dimension(H2)], i -> Random(Om) );
              u:= cf*Basis(H2);
              adun:= AdjointMatrix( Basis(L), u )^Dimension(L);
              if RankMat( adun ) + Dimension(H2) = Dimension(L) then
                 done:= true;
              fi;
           od;

           for i in [1..Dimension(L)] do
               k:= 0;
               done:= false;
               while not done do
                   cf[i]:=k;
                   u:= cf*Basis(H2);
                   adun:= AdjointMatrix( Basis(L), u )^Dimension(L);
                   if RankMat( adun ) + Dimension(H2) = Dimension(L) then
                      done:= true;
                   else
                      k:= k+1;
                   fi;
               od;
           od;

           bas:= Concatenation( Basis(H0), Basis(I) );
           B:= Basis(L,bas);
           cf:= Coefficients( B, u );
           y:= cf{[Dimension(H0)+1..Length(cf)]}*Basis(I);

           B:= Basis( I, List( Basis(I), x -> x*u ) );
           cf:= Coefficients( B, y );
           z:= cf*Basis(I);
           return Concatenation( [z], aa );
        fi;

end;

               
conjugatereductive:= function( L, R1, R2, N )

        # here L is an algebraic matrix Lie algebra (char 0),
        # R1, R2 are reductive subalgebras, and N is the nilpotent radical and
        # L = R1 + N = R2 + N. We compute a list of elements
        # x1,...,xs of N such that
        # exp(ad x1)...exp(ad xs)(R1) = R2

        local S1, S2, R, a, T1, T2, Lhat, H1, H2, b;

        S1:= LieDerivedSubalgebra( R1 );
        S2:= LieDerivedSubalgebra( R2 );
        R:= LieSolvableRadical( L );

        a:= conjugatesemsimparts( L, S1, S2, R );

        T1:= Subalgebra( L, List( Basis(LieCentre(R1)), x -> imexp( a, x ) ) );
        T2:= LieCentre( R2 );
        Lhat:= LieCentralizer( L, S2 );
        H1:= CartanSubalgebra( LieCentralizer( Lhat, T1 ) );
        H2:= CartanSubalgebra( LieCentralizer( Lhat, T2 ) );

        b:= conjCSAsolv( Lhat, H1, H2 );

        return Concatenation( b, [a] );

end;


specialCSA:= function( L, S, T, N, H )

        # here L is the Lie algebra of an algebraic matrix group, where
        # L = S + T + N is the standard decomposition. H is a CSA of L
        # we find x1,..,xk in N such that exp( ad x1 )...exp(ad xk )(H)
        # is a CSA containing H_s+T, where H_s is a CSA of S.

        local R, f, K, H0, B, vv, a, cf;

        R:= Ideal( L, Concatenation( Basis(T), Basis(N) ) );
        f:= NaturalHomomorphismByIdeal( L, R );
        K:= Range( f );
        H0:= Subalgebra( K, List( Basis(H), x -> Image( f, x ) ) );
        B:= Basis( L, Concatenation( Basis(S), Basis(R) ) );

        vv:= [ ];
        for a in Basis(H0) do
            cf:= Coefficients( B, PreImagesRepresentative( f, a ) );
            cf:= cf{[1..Dimension(S)]};
            Add( vv, cf*Basis(S) );
        od;
        Append( vv, Basis(T) );
        H0:= CartanSubalgebra( LieCentralizer( L, Subalgebra( L, vv ) ) );

        B:= Subalgebra( L, Concatenation( Basis(H0), Basis(R) ) );
        
        return conjCSAsolv( B, H, H0 );
end;

projection:= function( L, S, T, N, g )

        # Input is the same as in the previous function; we consider the
        # group G^r corresponding to S+T, we compute the image of g in G
        # under the projection G --> G^r. 

        local jd, s, u, u1, u2, lgu, i, B, cf, mm, M, BM, aa, sol, C, H,
              xx, exps, x, y, e, t, h, vv1;

        if Dimension(N) = 0 then return g; fi;
        
        jd:= JordanDecomposition( g );
        s:= jd[1]; u:= s^0+s^-1*jd[2];

        u1:= u-u^0;
        u2:= u1;
        lgu:= Zero(LeftActingDomain(L))*u;
        i:= 1;
        while not IsZero(u2) do
            lgu:= lgu+(-1)^(i-1)*u2/i;
            u2:= u2*u1; i:= i+1;
        od;

        vv1:= Concatenation( Basis(S), Basis(T) );
        B:= Basis(L, Concatenation( vv1, Basis(N) ) );
        cf:= Coefficients( B, LieObject(lgu) );
        u:= cf{[1..Length(vv1)]}*vv1;
        u:= u![1];

        # exponentiate...

        u1:= u^0;
        i:=1;
        u2:= u;
        while not IsZero(u2) do
           u1:= u1+u2;
           u2:= u2*u/(i+1);
           i:= i+1;
        od;

        mm:= List( Basis(L), x -> x![1] );
        M:= VectorSpace( LeftActingDomain(L), mm );
        BM:= Basis( M, mm );
        aa:= List( mm, x -> Coefficients( BM, s*x*s^-1 ) );
        for i in [1..Length(aa)] do aa[i][i]:= aa[i][i]-1; od;

        sol:= NullspaceMat( aa );
        C:= Subalgebra( L, List( sol, x -> LieObject( x*mm ) ) );

        H:= CartanSubalgebra( C );

        xx:= specialCSA( L, S, T, N, H );

        exps:= [ ];
        for x in xx do
            y:= x![1];
            e:= y^0;
            t:= y;
            i:= 1;
            while not IsZero(t) do
                e:= e+t;
                t:= t*y/(i+1);
                i:= i+1;
            od;
            Add( exps, e );
        od;

        h:= Product( exps );
        return h*s*h^-1*u1;
end;
